# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

from tapqir.imscroll.glimpse_reader import GlimpseDataset, read_glimpse

__all__ = [
    "GlimpseDataset",
    "read_glimpse",
]
